import os

from .exceptions import ArgcompleteException
from .io import debug
from .packages import _shlex


def split_line(line, point=None):
    if point is None:
        point = len(line)
    line = line[:point]
    lexer = _shlex.shlex(line, posix=True)
    lexer.whitespace_split = True
    lexer.wordbreaks = os.environ.get("_ARGCOMPLETE_COMP_WORDBREAKS", "")
    words = []

    def split_word(word):
        # TODO: make this less ugly
        point_in_word = len(word) + point - lexer.instream.tell()
        if isinstance(lexer.state, (str, bytes)) and lexer.state in lexer.whitespace:
            point_in_word += 1
        if point_in_word > len(word):
            debug("In trailing whitespace")
            words.append(word)
            word = ""
        prefix, suffix = word[:point_in_word], word[point_in_word:]
        prequote = ""
        # posix
        if lexer.state is not None and lexer.state in lexer.quotes:
            prequote = lexer.state
        # non-posix
        # if len(prefix) > 0 and prefix[0] in lexer.quotes:
        #    prequote, prefix = prefix[0], prefix[1:]

        return prequote, prefix, suffix, words, lexer.last_wordbreak_pos

    while True:
        try:
            word = lexer.get_token()
            if word == lexer.eof:
                # TODO: check if this is ever unsafe
                # raise ArgcompleteException("Unexpected end of input")
                return "", "", "", words, None
            if lexer.instream.tell() >= point:
                debug("word", word, "split, lexer state: '{s}'".format(s=lexer.state))
                return split_word(word)
            words.append(word)
        except ValueError:
            debug("word", lexer.token, "split (lexer stopped, state: '{s}')".format(s=lexer.state))
            if lexer.instream.tell() >= point:
                return split_word(lexer.token)
            else:
                msg = (
                    "Unexpected internal state. "
                    "Please report this bug at https://github.com/kislyuk/argcomplete/issues."
                )
                raise ArgcompleteException(msg)



"""
Utility for locating a module (or package's __main__.py) with a given name
and verifying it contains the PYTHON_ARGCOMPLETE_OK marker.

The module name should be specified in a form usable with `python -m`.

Intended to be invoked by argcomplete's global completion function.
"""

import os
import sys
import tokenize
from importlib.util import find_spec


class ArgcompleteMarkerNotFound(RuntimeError):
    pass


def find(name, return_package=False):
    names = name.split(".")
    # Look for the first importlib ModuleSpec that has `origin` set, indicating it's not a namespace package.
    for package_name_boundary in range(len(names)):
        spec = find_spec(".".join(names[: package_name_boundary + 1]))
        if spec is not None and spec.origin is not None:
            break

    if spec is None:
        raise ArgcompleteMarkerNotFound('no module named "{}"'.format(names[0]))
    if not spec.has_location:
        raise ArgcompleteMarkerNotFound("cannot locate file")
    if spec.submodule_search_locations is None:
        if len(names) != 1:
            raise ArgcompleteMarkerNotFound("{} is not a package".format(names[0]))
        return spec.origin
    if len(spec.submodule_search_locations) != 1:
        raise ArgcompleteMarkerNotFound("expecting one search location")
    path = os.path.join(spec.submodule_search_locations[0], *names[package_name_boundary + 1 :])
    if os.path.isdir(path):
        filename = "__main__.py"
        if return_package:
            filename = "__init__.py"
        return os.path.join(path, filename)
    else:
        return path + ".py"


def main():
    try:
        name = sys.argv[1]
    except IndexError:
        raise ArgcompleteMarkerNotFound("missing argument on the command line")

    filename = find(name)

    try:
        fp = tokenize.open(filename)
    except OSError:
        raise ArgcompleteMarkerNotFound("cannot open file")

    with fp:
        head = fp.read(1024)

    if "PYTHON_ARGCOMPLETE_OK" not in head:
        raise ArgcompleteMarkerNotFound("marker not found")


if __name__ == "__main__":
    try:
        main()
    except ArgcompleteMarkerNotFound as e:
        sys.exit(str(e))



import contextlib
import os
import sys

_DEBUG = "_ARC_DEBUG" in os.environ

debug_stream = sys.stderr


def debug(*args):
    if _DEBUG:
        print(file=debug_stream, *args)


@contextlib.contextmanager
def mute_stdout():
    stdout = sys.stdout
    sys.stdout = open(os.devnull, "w")
    try:
        yield
    finally:
        sys.stdout = stdout


@contextlib.contextmanager
def mute_stderr():
    stderr = sys.stderr
    sys.stderr = open(os.devnull, "w")
    try:
        yield
    finally:
        sys.stderr.close()
        sys.stderr = stderr


def warn(*args):
    """
    Prints **args** to standard error when running completions. This will interrupt the user's command line interaction;
    use it to indicate an error condition that is preventing your completer from working.
    """
    print(file=debug_stream)
    print(file=debug_stream, *args)



#!/usr/bin/env python3

# Copyright 2012-2023, Andrey Kislyuk and argcomplete contributors.
# Licensed under the Apache License. See https://github.com/kislyuk/argcomplete for more info.

"""
This script is part of the Python argcomplete package (https://github.com/kislyuk/argcomplete).
It is used to check if an EASY-INSTALL-SCRIPT wrapper redirects to a script that contains the string
"PYTHON_ARGCOMPLETE_OK". If you have enabled global completion in argcomplete, the completion hook will run it every
time you press <TAB> in your shell.

Usage:
    python-argcomplete-check-easy-install-script <input executable file>
"""

import sys

# PEP 366
__package__ = "argcomplete.scripts"


def main():
    if len(sys.argv) != 2:
        sys.exit(__doc__)

    sys.tracebacklimit = 0

    with open(sys.argv[1]) as fh:
        line1, head = fh.read(1024).split("\n", 1)[:2]
        if line1.startswith("#") and ("py" in line1 or "Py" in line1):
            import re

            lines = head.split("\n", 12)
            for line in lines:
                if line.startswith("# EASY-INSTALL-SCRIPT"):
                    import pkg_resources  # type: ignore

                    re_match = re.match("# EASY-INSTALL-SCRIPT: '(.+)','(.+)'", line)
                    assert re_match is not None
                    dist, script = re_match.groups()
                    if "PYTHON_ARGCOMPLETE_OK" in pkg_resources.get_distribution(dist).get_metadata(
                        "scripts/" + script
                    ):
                        return 0
                elif line.startswith("# EASY-INSTALL-ENTRY-SCRIPT"):
                    re_match = re.match("# EASY-INSTALL-ENTRY-SCRIPT: '(.+)','(.+)','(.+)'", line)
                    assert re_match is not None
                    dist, group, name = re_match.groups()
                    import pkgutil

                    import pkg_resources  # type: ignore

                    entry_point_info = pkg_resources.get_distribution(dist).get_entry_info(group, name)
                    assert entry_point_info is not None
                    module_name = entry_point_info.module_name
                    with open(pkgutil.get_loader(module_name).get_filename()) as mod_fh:  # type: ignore
                        if "PYTHON_ARGCOMPLETE_OK" in mod_fh.read(1024):
                            return 0
                elif line.startswith("# EASY-INSTALL-DEV-SCRIPT"):
                    for line2 in lines:
                        if line2.startswith("__file__"):
                            re_match = re.match("__file__ = '(.+)'", line2)
                            assert re_match is not None
                            filename = re_match.group(1)
                            with open(filename) as mod_fh:
                                if "PYTHON_ARGCOMPLETE_OK" in mod_fh.read(1024):
                                    return 0
                elif line.startswith("# PBR Generated"):
                    re_match = re.search("from (.*) import", head)
                    assert re_match is not None
                    module = re_match.groups()[0]
                    import pkgutil

                    import pkg_resources  # type: ignore

                    with open(pkgutil.get_loader(module).get_filename()) as mod_fh:  # type: ignore
                        if "PYTHON_ARGCOMPLETE_OK" in mod_fh.read(1024):
                            return 0

    return 1


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python3
# PYTHON_ARGCOMPLETE_OK

# Copyright 2012-2023, Andrey Kislyuk and argcomplete contributors.
# Licensed under the Apache License. See https://github.com/kislyuk/argcomplete for more info.

"""
Register a Python executable for use with the argcomplete module.

To perform the registration, source the output of this script in your bash shell
(quote the output to avoid interpolation).

Example:

    $ eval "$(register-python-argcomplete my-favorite-script.py)"

For Tcsh

    $ eval `register-python-argcomplete --shell tcsh my-favorite-script.py`

For Fish

    $ register-python-argcomplete --shell fish my-favourite-script.py > ~/.config/fish/my-favourite-script.py.fish
"""

import argparse
import sys

import argcomplete

# PEP 366
__package__ = "argcomplete.scripts"


def main():
    parser = argparse.ArgumentParser(description=__doc__, formatter_class=argparse.RawDescriptionHelpFormatter)

    parser.add_argument(
        "--no-defaults",
        dest="use_defaults",
        action="store_false",
        default=True,
        help="when no matches are generated, do not fallback to readline's default completion (affects bash only)",
    )
    parser.add_argument(
        "--complete-arguments",
        nargs=argparse.REMAINDER,
        help="arguments to call complete with; use of this option discards default options (affects bash only)",
    )
    parser.add_argument(
        "-s",
        "--shell",
        choices=("bash", "zsh", "tcsh", "fish", "powershell"),
        default="bash",
        help="output code for the specified shell",
    )
    parser.add_argument(
        "-e", "--external-argcomplete-script", help="external argcomplete script for auto completion of the executable"
    )

    parser.add_argument("executable", nargs="+", help="executable to completed (when invoked by exactly this name)")

    argcomplete.autocomplete(parser)

    if len(sys.argv) == 1:
        parser.print_help()
        sys.exit(1)

    args = parser.parse_args()

    sys.stdout.write(
        argcomplete.shellcode(
            args.executable, args.use_defaults, args.shell, args.complete_arguments, args.external_argcomplete_script
        )
    )


if __name__ == "__main__":
    sys.exit(main())
